from enum import Enum
from itertools import product
from typing import Literal

import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.gridspec import GridSpec
from pulp import PULP_CBC_CMD, LpInteger, LpMaximize, LpMinimize, LpProblem, LpVariable, lpSum
from shapely import LineString

from ..models.region import Town
from ..models.service_type import ServiceType
from .base_method import BaseMethod


PROVISION_COLUMN = "provision"
DEMAND_COLUMN = "demand"
DEMAND_WITHIN_COLUMN = "demand_within"
DEMAND_WITHOUT_COLUMN = "demand_without"
DEMAND_LEFT_COLUMN = "demand_left"
CAPACITY_COLUMN = "capacity"
CAPACITY_LEFT_COLUMN = "capacity_left"

PLOT_KWARGS = {"column": PROVISION_COLUMN, "cmap": "RdYlGn", "vmin": 0, "vmax": 1, "legend": True}


class ProvisionMethod(Enum):
    GRAVITATIONAL = "gravitational"
    LINEAR = "linear"


class Provision(BaseMethod):
    """Class provides methods for service type provision assessment"""

    @staticmethod
    def _plot_provision(ax, gdf):
        gdf.sort_values(by=PROVISION_COLUMN).plot(ax=ax, zorder=1, **PLOT_KWARGS)

    def _plot_districts(self, ax, gdf):
        self._plot_provision(ax, gdf)
        ax.set_title("Districts")

    def _plot_settlements(self, ax, gdf):
        self._plot_provision(ax, gdf)
        ax.set_title("Settlements")

    def _plot_towns(self, ax, service_type, prov_gdf, links_gdf):
        service_type_name = service_type.name
        total_provision = self.total_provision(prov_gdf)
        links_gdf.sort_values(by=DEMAND_COLUMN).plot(
            ax=ax,
            zorder=0,
            column=DEMAND_COLUMN,
            cmap="cool",
            alpha=1,
            vmin=0,
            vmax=200,
        )
        self._plot_provision(ax, prov_gdf)
        ax.set_title(f"Provision {service_type_name}: {total_provision: .3f}")

    def _aggregate_unit(self, groupby):
        res = groupby.agg(
            {
                CAPACITY_COLUMN: "sum",
                CAPACITY_LEFT_COLUMN: "sum",
                DEMAND_COLUMN: "sum",
                DEMAND_LEFT_COLUMN: "sum",
                DEMAND_WITHIN_COLUMN: "sum",
                DEMAND_WITHOUT_COLUMN: "sum",
            }
        )
        res[PROVISION_COLUMN] = res[DEMAND_WITHIN_COLUMN] / res[DEMAND_COLUMN]
        return res

    def _aggregate_units(self, prov_gdf) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        settlements_gdf = self._aggregate_unit(prov_gdf.groupby("settlement_name"))
        districts_gdf = self._aggregate_unit(prov_gdf.groupby("district_name"))
        settlements_gdf = self.region.settlements.merge(settlements_gdf, left_on="name", right_index=True)
        districts_gdf = self.region.districts.merge(districts_gdf, left_on="name", right_index=True)
        return settlements_gdf, districts_gdf

    def plot(
        self,
        service_type: str | ServiceType,
        districts_gdf: gpd.GeoDataFrame,
        settlements_gdf: gpd.GeoDataFrame,
        towns_gdf: gpd.GeoDataFrame,
        links_gdf: gpd.GeoDataFrame,
        figsize=(15, 15),
    ):
        """Visualizes provision assessment results"""
        if not isinstance(service_type, ServiceType):
            service_type = self.region[service_type]
        fig = plt.figure(figsize=figsize)
        grid = GridSpec(3, 2)
        axes = {
            "towns": fig.add_subplot(grid[0:2, 0:2]),
            "settlements": fig.add_subplot(grid[2, 0]),
            "districts": fig.add_subplot(grid[2, 1]),
        }
        for ax in axes.values():
            ax.set_axis_off()
        self._plot_towns(axes["towns"], service_type, towns_gdf, links_gdf)
        self._plot_settlements(axes["settlements"], settlements_gdf)
        self._plot_districts(axes["districts"], districts_gdf)

    def _get_towns_gdf(self, service_type: ServiceType) -> gpd.GeoDataFrame:
        """Returns towns gdf for provision assessment"""
        capacity_column = f"capacity_{service_type.name}"
        gdf = self.region.get_towns_gdf()[
            ["geometry", "town_name", "settlement_name", "district_name", "population", capacity_column]
        ].fillna(0)
        gdf["population"] = gdf["population"].apply(service_type.calculate_in_need)
        gdf = gdf.rename(columns={"population": DEMAND_COLUMN, capacity_column: CAPACITY_COLUMN})
        gdf[CAPACITY_LEFT_COLUMN] = gdf[CAPACITY_COLUMN]
        gdf[DEMAND_LEFT_COLUMN] = gdf[DEMAND_COLUMN]
        gdf[DEMAND_WITHIN_COLUMN] = 0
        gdf[DEMAND_WITHOUT_COLUMN] = 0
        return gdf

    @classmethod
    def total_provision(cls, gdf: gpd.GeoDataFrame):
        return gdf[DEMAND_WITHIN_COLUMN].sum() / gdf[DEMAND_COLUMN].sum()

    def _link_to_linestring(self, dict):
        town_from = self.region[int(dict["from"])]
        town_to = self.region[int(dict["to"])]
        return LineString([town_from.geometry, town_to.geometry])

    def calculate(
        self,
        service_type: ServiceType | str,
        update_df: pd.DataFrame | None = None,
        method: ProvisionMethod = ProvisionMethod.GRAVITATIONAL,
        self_supply: bool = True,
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame, gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """Provision assessment using certain method for the region and
        service type, can be used with certain updated capacity and population info df"""

        if not isinstance(service_type, ServiceType):
            service_type: ServiceType = self.region[service_type]
        towns_gdf = self._get_towns_gdf(service_type)

        if update_df is not None:
            towns_gdf = towns_gdf.join(update_df)
            towns_gdf[update_df.columns] = towns_gdf[update_df.columns].fillna(0)
            if not "population" in towns_gdf:
                towns_gdf["population"] = 0
            towns_gdf[DEMAND_COLUMN] += towns_gdf["population"].apply(service_type.calculate_in_need)
            towns_gdf[DEMAND_LEFT_COLUMN] = towns_gdf[DEMAND_COLUMN]
            if not service_type.name in towns_gdf:
                towns_gdf[service_type.name] = 0
            towns_gdf[CAPACITY_COLUMN] += towns_gdf[service_type.name]
            towns_gdf[CAPACITY_LEFT_COLUMN] += towns_gdf[service_type.name]

        if self_supply:
            supply: pd.Series = towns_gdf.apply(lambda x: min(x[DEMAND_COLUMN], x[CAPACITY_COLUMN]), axis=1)
            towns_gdf[DEMAND_WITHIN_COLUMN] += supply
            towns_gdf[DEMAND_LEFT_COLUMN] -= supply
            towns_gdf[CAPACITY_LEFT_COLUMN] -= supply

        towns_gdf, links = self._tp_provision(towns_gdf, service_type, method)

        for link in links:
            link["geometry"] = self._link_to_linestring(link)
        links_gdf = gpd.GeoDataFrame(links, columns=["geometry", "from", "to", "demand", "within"]).set_crs(
            self.region.crs
        )

        towns_gdf[PROVISION_COLUMN] = towns_gdf[DEMAND_WITHIN_COLUMN] / towns_gdf[DEMAND_COLUMN]
        settlements_gdf, districts_gdf = self._aggregate_units(towns_gdf)
        return districts_gdf, settlements_gdf, towns_gdf, links_gdf

    def _tp_provision(
        self, gdf: gpd.GeoDataFrame, service_type: ServiceType, method: ProvisionMethod
    ) -> tuple[gpd.GeoDataFrame, list[dict]]:
        """Transport problem assessment method"""

        links = []
        # if delta is not 0, we add a dummy city block
        # delta = gdf[DEMAND_COLUMN].sum() - gdf[CAPACITY_COLUMN].sum()
        # fictive_demand = None
        # fictive_capacity = None
        # fictive_town_id = gdf.index.max() + 1
        # if delta > 0:
        #     fictive_capacity = fictive_town_id
        #     gdf.loc[fictive_capacity, CAPACITY_COLUMN] = delta
        #     gdf.loc[fictive_capacity, CAPACITY_LEFT_COLUMN] = delta
        # if delta < 0:
        #     fictive_demand = fictive_town_id
        #     gdf.loc[fictive_demand, DEMAND_COLUMN] = -delta
        #     gdf.loc[fictive_demand, DEMAND_LEFT_COLUMN] = -delta

        def _get_distance(id1: int, id2: int):
            # if id1 == fictive_town_id or id2 == fictive_town_id:
            # return 0
            distance = self.region[id1, id2]
            return distance if distance > 1 else 1

        def _get_weight(id1: int, id2: int):
            distance = _get_distance(id1, id2)
            if method == ProvisionMethod.GRAVITATIONAL:
                return 1 / (distance * distance)
            return 1 / distance

        demand = gdf.loc[gdf[DEMAND_LEFT_COLUMN] > 0]
        capacity = gdf.loc[gdf[CAPACITY_LEFT_COLUMN] > 0]

        prob = LpProblem("Provision", LpMaximize)
        products = []
        for i in demand.index:
            for j in capacity.index:
                if _get_distance(i, j) <= service_type.accessibility * 3:
                    products.append((i, j))
        x = LpVariable.dicts("Route", products, 0, None, cat=LpInteger)
        prob += lpSum(_get_weight(n, m) * x[n, m] for n, m in products)
        for n in demand.index:
            capacity_products = [tpl[1] for tpl in filter(lambda tpl: tpl[0] == n, products)]
            prob += lpSum(x[n, m] for m in capacity_products) <= demand.loc[n, DEMAND_LEFT_COLUMN]
        for m in capacity.index:
            demand_products = [tpl[0] for tpl in filter(lambda tpl: tpl[1] == m, products)]
            prob += lpSum(x[n, m] for n in demand_products) <= capacity.loc[m, CAPACITY_LEFT_COLUMN]
        prob.solve(PULP_CBC_CMD(msg=False))

        for var in prob.variables():
            value = var.value()
            name = var.name.replace("(", "").replace(")", "").replace(",", "").split("_")
            if name[2] == "dummy":
                continue
            a = int(name[1])
            b = int(name[2])
            distance = _get_distance(a, b)
            if value > 0:
                # if a != fictive_demand and b != fictive_capacity:
                if distance <= service_type.accessibility:
                    gdf.loc[a, DEMAND_WITHIN_COLUMN] += value
                else:
                    gdf.loc[a, DEMAND_WITHOUT_COLUMN] += value
                gdf.loc[a, DEMAND_LEFT_COLUMN] -= value
                gdf.loc[b, CAPACITY_LEFT_COLUMN] -= value
                links.append({"from": a, "to": b, "demand": value, "within": distance <= service_type.accessibility})

        # if fictive_town_id is not None:
        #     gdf = gdf.drop(labels=[fictive_town_id])

        return gdf, links



COMFORT_SERVICE_TYPES = {
    "education": [
        {
            "name": "art_school",
            "name_ru": "школа искусств",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 8,
            "osm_tags": {"amenity": "music_school"},
        },
    ],
    "healthcare": [
        {
            "name": "city_hospital",
            "name_ru": "городская больница (д/в)",
            "weight": 0.5,
            "accessibility": 60,
            "demand": 9,
            "osm_tags": {"amenity": "hospital"},
        },
        {
            "name": "commercial_clinic",
            "name_ru": "коммерческая клиника",
            "weight": 0.3,
            "accessibility": 60,
            "demand": 9,
            "osm_tags": {"amenity": "clinic"},
        },
    ],
    "commerce": [
        {
            "name": "hypermarket",
            "name_ru": "ТРК / гипермаркет",
            "weight": 0.5,
            "accessibility": 30,
            "demand": 10,
            "osm_tags": {"shop": "mall"},
        },
        {
            "name": "specialized_store",
            "name_ru": "профильный магазин",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 10,
            "osm_tags": {"shop": "*"},
        },
    ],
    "catering": [
        {
            "name": "food_court",
            "name_ru": "фудкорт",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 105,
            "osm_tags": {"amenity": "food_court"},
        },
    ],
    "leisure": [
        {
            "name": "library",
            "name_ru": "медиатека / библиотека",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 100,
            "osm_tags": {"amenity": "library"},
        },
        {
            "name": "cult_object",
            "name_ru": "культовый объект",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 100,
            "osm_tags": {"building": "church"},
        },
    ],
    "recreation": [
        {
            "name": "amusement_park",
            "name_ru": "парк развлечений",
            "weight": 0.2,
            "accessibility": 60,
            "demand": 100,
            "osm_tags": {"tourism": "theme_park"},
        },
        {
            "name": "dog_park",
            "name_ru": "площадка для выгула собак",
            "weight": 0.2,
            "accessibility": 15,
            "demand": 53,
            "osm_tags": {"tourism": "theme_park"},
        },
    ],
    "sport": [
        {
            "name": "swimming_pool",
            "name_ru": "ФОК / бассейн",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 8,
            "osm_tags": {"leisure": "swimming_pool"},
        },
    ],
    "service": [],
    "transport": [
        {
            "name": "railway_station",
            "name_ru": "ЖД станция",
            "weight": 0.4,
            "accessibility": 15,
            "demand": 50,
            "osm_tags": {"railway": "station"},
        },
        {
            "name": "car_service",
            "name_ru": "автосалон / автосервис",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 50,
            "osm_tags": {"shop": "car_repair"},
        },
    ],
    "safeness": [
        {
            "name": "fire_station",
            "name_ru": "пожарное депо",
            "weight": 0.5,
            "accessibility": 15,
            "demand": 20,
            "osm_tags": {"amenity": "fire_station"},
        },
    ],
}



from abc import ABC

from pydantic import BaseModel, InstanceOf

from ..models.region import Region


class BaseMethod(ABC, BaseModel):
    """BaseMethod class required for methods implementation"""

    region: InstanceOf[Region]



import math
from enum import Enum

import pandas as pd
from pydantic import BaseModel, Field, field_validator


class ServiceInfrastructure(Enum):
    EDUCATION = "education"
    HEALTHCARE = "healthcare"
    COMMERCE = "commerce"
    CATERING = "catering"
    LEISURE = "leisure"
    RECREATION = "recreation"
    SPORT = "sport"
    SERVICE = "service"
    TRANSPORT = "transport"
    SAFENESS = "safeness"


class ServiceCategory(Enum):
    BASIC = "basic"
    BASIC_PLUS = "basic_plus"
    COMFORT = "comfort"


class ServiceType(BaseModel):
    """Represents service type entity, such as schools and its parameters overall"""

    category: ServiceCategory
    infrastructure: ServiceInfrastructure
    name: str
    name_ru: str
    weight: float = Field(gt=0)
    accessibility: int = Field(gt=0)
    demand: int = Field(gt=0)
    osm_tags: dict
    # bricks: list[ServiceBrick] = []

    # @field_validator("bricks", mode="before")
    # def validate_bricks(value):
    #     bricks = [sb if isinstance(sb, ServiceBrick) else ServiceBrick(**sb) for sb in value]
    #     return bricks

    def calculate_in_need(self, population: int) -> int:
        """Calculate how many people in the given population are in need by this service type"""
        return math.ceil(population / 1000 * self.demand)

    def to_dict(self):
        return {
            "category": self.category.name,
            "infrastructure": self.infrastructure.name,
            "name": self.name,
            "name_ru": self.name_ru,
            "weight": self.weight,
            "accessibility": self.accessibility,
            "demand": self.demand,
        }

    @staticmethod
    def to_df(iterable):
        if isinstance(iterable, dict):
            iterable = iterable.values()
        return pd.DataFrame([st.to_dict() for st in iterable])

    def __hash__(self):
        """Make service type hashable to use it as a key"""
        return hash(self.name)

    def __str__(self):
        return str.join("\n", [f"{key.ljust(15)}: {value}" for key, value in self.to_dict().items()])



BASIC_SERVICE_TYPES = {
    "education": [
        {
            "name": "kindergarten",
            "name_ru": "детский сад",
            "weight": 0.2,
            "accessibility": 7,
            "demand": 61,
            "osm_tags": {"amenity": "kindergarten"},
        },
        {
            "name": "school",
            "name_ru": "школа",
            "weight": 0.2,
            "accessibility": 15,
            "demand": 120,
            "osm_tags": {"amenity": "school"},
        },
    ],
    "healthcare": [
        {
            "name": "health_center__dispensary",
            "name_ru": "ФАП / амбулатория",
            "weight": 0.2,
            "accessibility": 10,
            "demand": 13,
            "osm_tags": {"amenity": "clinic"},
        },
        {
            "name": "pharmacy",
            "name_ru": "аптека",
            "weight": 0.2,
            "accessibility": 15,
            "demand": 15,
            "osm_tags": {"amenity": "pharmacy"},
        },
    ],
    "commerce": [
        {
            "name": "convenience",
            "name_ru": "продуктовый магазин",
            "weight": 0.2,
            "accessibility": 5,
            "demand": 180,
            "osm_tags": {"shop": "supermarket"},
        },
        {
            "name": "houseware",
            "name_ru": "хозяйственный магазин",
            "weight": 0.2,
            "accessibility": 5,
            "demand": 180,
            "osm_tags": {"shop": "houseware"},
        },
    ],
    "catering": [
        {
            "name": "cafe__coffee",
            "name_ru": "кафе / кофейня",
            "weight": 0.15,
            "accessibility": 15,
            "demand": 25,
            "osm_tags": {"amenity": "cafe"},
        },
    ],
    "leisure": [
        {
            "name": "universal_hall",
            "name_ru": "универсальный зал",
            "weight": 0.1,
            "accessibility": 15,
            "demand": 180,
            "osm_tags": {"leisure": "sports_hall"},
        },
    ],
    "recreation": [
        {
            "name": "playground",
            "name_ru": "детская площадка",
            "weight": 0.2,
            "accessibility": 4,
            "demand": 2,
            "osm_tags": {"leisure": "playground"},
        },
        {
            "name": "square__boulevard__forest_park",
            "name_ru": "сквер / бульвар / лесопарк",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 150,
            "osm_tags": {"leisure": "park"},
        },
    ],
    "sport": [
        {
            "name": "workout__school_gym",
            "name_ru": "воркаут / школьный спортзал",
            "weight": 0.15,
            "accessibility": 45,
            "demand": 10,
            "osm_tags": {"leisure": "fitness_station"},
        },
    ],
    "service": [
        {
            "name": "delivery_point__post_office",
            "name_ru": "пункт доставки / почтовое отделение",
            "weight": 0.1,
            "accessibility": 15,
            "demand": 10,
            "osm_tags": {"shop": "outpost", "amenity": "post_office"},
        },
        {
            "name": "hairdresser__beauty",
            "name_ru": "парикмахерская / салон красоты",
            "weight": 0.1,
            "accessibility": 15,
            "demand": 10,
            "osm_tags": {"shop": ["beauty", "hairdresser"], "beauty": ["nails", "tanning", "spa", "massage"]},
        },
    ],
    "transport": [
        {
            "name": "public_transport_stop",
            "name_ru": "остановка ОТ",
            "weight": 0.2,
            "accessibility": 10,
            "demand": 500,
            "osm_tags": {"highway": "bus_stop"},
        },
        {
            "name": "parking",
            "name_ru": "парковка",
            "weight": 0.2,
            "accessibility": 7,
            "demand": 378,
            "osm_tags": {"amenity": "parking"},
        },
    ],
    "safeness": [
        {
            "name": "local_police",
            "name_ru": "участковый пункт полиции",
            "weight": 0.4,
            "accessibility": 10,
            "demand": 5,
            "osm_tags": {"amenity": "police"},
        },
    ],
}



import pickle
from functools import singledispatchmethod

import geopandas as gpd
import osmnx as ox
import pandas as pd
import shapely

from ..utils import SERVICE_TYPES
from .service_type import ServiceType
from .territory import Territory
from .town import Town


DISTRICTS_PLOT_COLOR = "#e40e37"
SETTLEMENTS_PLOT_COLOR = "#ddd"
TOWNS_PLOT_COLOR = "#777"
SERVICES_PLOT_COLOR = "#aaa"
TERRITORIES_PLOT_COLOR = "#29392C"


class Region:

    def __init__(
        self,
        districts: gpd.GeoDataFrame,
        settlements: gpd.GeoDataFrame,
        towns: gpd.GeoDataFrame,
        accessibility_matrix: pd.DataFrame,
        territories: gpd.GeoDataFrame | None = None,
    ):

        districts = self.validate_districts(districts)
        settlements = self.validate_settlements(settlements)
        towns = self.validate_towns(towns)
        accessibility_matrix = self.validate_accessibility_matrix(accessibility_matrix)

        assert (
            accessibility_matrix.index == towns.index
        ).all(), "Accessibility matrix indices and towns indices don't match"
        assert districts.crs == settlements.crs == towns.crs, "CRS should match everywhere"

        self.crs = towns.crs
        self.districts = districts
        self.settlements = settlements
        self._towns = Town.from_gdf(towns)
        self.accessibility_matrix = accessibility_matrix
        if territories is None:
            self._territories = {}
        else:
            assert territories.crs == towns.crs, "Territories CRS should match towns CRS"
            self._territories = Territory.from_gdf(territories)

        service_types = {}
        for service_type_dict in SERVICE_TYPES:
            service_types[service_type_dict["name"]] = ServiceType(**service_type_dict)
        self._service_types = service_types

    @staticmethod
    def validate_districts(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        assert isinstance(gdf, gpd.GeoDataFrame), "Districts should be instance of gpd.GeoDataFrame"
        assert gdf.geom_type.isin(
            ["Polygon", "MultiPolygon"]
        ).all(), "District geometry should be Polygon or MultiPolygon"
        assert pd.api.types.is_string_dtype(gdf["name"]), "District name should be str"
        return gdf[["geometry", "name"]]

    @staticmethod
    def validate_settlements(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        assert isinstance(gdf, gpd.GeoDataFrame), "Settlements should be instance of gpd.GeoDataFrame"
        assert gdf.geom_type.isin(
            ["Polygon", "MultiPolygon"]
        ).all(), "Settlement geometry should be Polygon or MultiPolygon"
        assert pd.api.types.is_string_dtype(gdf["name"]), "Settlement name should be str"
        return gdf[["geometry", "name"]]

    @staticmethod
    def validate_towns(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        assert isinstance(gdf, gpd.GeoDataFrame), "Towns should be instance of gpd.GeoDataFrame"
        return gdf

    @staticmethod
    def validate_accessibility_matrix(df: pd.DataFrame) -> pd.DataFrame:
        assert pd.api.types.is_float_dtype(df.values), "Accessibility matrix values should be float"
        assert (df.values >= 0).all(), "Accessibility matrix values should be greater or equal 0"
        assert (df.index == df.columns).all(), "Accessibility matrix indices and columns don't match"
        return df

    @staticmethod
    def validate_services(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        assert isinstance(gdf, gpd.GeoDataFrame), "Services should be instance of gpd.GeoDataFrame"
        return gdf[["geometry", "town_id", "capacity"]]

    def plot(self, figsize=(15, 15)):
        sett_to_dist = self.settlements.copy()
        sett_to_dist.geometry = sett_to_dist.representative_point()
        sett_to_dist = sett_to_dist.sjoin(self.districts).rename(
            columns={"name_right": "district_name", "index_right": "district_id", "name_left": "settlement_name"}
        )
        sett_to_dist.geometry = self.settlements.geometry
        ax = sett_to_dist.plot(facecolor="none", edgecolor=SETTLEMENTS_PLOT_COLOR, figsize=figsize)
        self.get_services_gdf().plot(ax=ax, markersize=1, color=SERVICES_PLOT_COLOR)
        self.get_towns_gdf().plot(ax=ax, markersize=2, color=TOWNS_PLOT_COLOR)
        self.districts.plot(ax=ax, facecolor="none", edgecolor=DISTRICTS_PLOT_COLOR)
        territories_gdf = self.get_territories_gdf()
        territories_gdf.geometry = territories_gdf.representative_point()
        territories_gdf.plot(ax=ax, marker="*", markersize=60, color=TERRITORIES_PLOT_COLOR)
        ax.set_axis_off()

    def get_territory(self, territory_id: int):
        if not territory_id in self._territories:
            raise KeyError(f"Can't find territory with such id: {territory_id}")
        return self._territories[territory_id]

    @property
    def towns(self) -> list[Town]:
        return self._towns.values()

    @property
    def territories(self) -> list[Territory]:
        return self._territories.values()

    @property
    def service_types(self) -> list[ServiceType]:
        return self._service_types.values()

    @property
    def geometry(self) -> shapely.Polygon | shapely.MultiPolygon:
        return self.districts.to_crs(4326).unary_union

    def match_services_towns(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        assert gdf.crs == self.crs, "Services GeoDataFrame CRS should match region CRS"
        gdf = gdf.copy()
        towns_gdf = self.get_towns_gdf()[["geometry", "population"]]

        def get_closest_city(service_i):
            # service_gdf = gdf[gdf.index == service_i]
            sjoin = towns_gdf.copy()
            sjoin["distance"] = towns_gdf.geometry.apply(lambda g: shapely.distance(g, gdf.loc[service_i, "geometry"]))
            sjoin["weight"] = sjoin["population"] / sjoin["distance"] / sjoin["distance"]
            return sjoin["weight"].idxmax()

        gdf["town_id"] = gdf.apply(lambda s: get_closest_city(s.name), axis=1)
        return gdf

    def update_services(self, service_type: ServiceType | str, gdf: gpd.GeoDataFrame):
        """Update capacities in towns of certain service_type"""
        gdf = self.validate_services(gdf)
        assert gdf.crs == self.crs, "Services GeoDataFrame CRS should match region CRS"
        if not isinstance(service_type, ServiceType):
            service_type = self[service_type]
        # reset services of towns
        for town in self.towns:
            town.update_services(service_type)
        # spatial join blocks and services and update related blocks info
        groups = gdf.groupby("town_id")
        for town_id, services_gdf in groups:
            self[town_id].update_services(service_type, services_gdf)

    def get_territories_gdf(self) -> gpd.GeoDataFrame:
        data = [territory.to_dict() for territory in self.territories]
        return gpd.GeoDataFrame(data, crs=self.crs).set_index("id", drop=True)

    def get_services_gdf(self) -> gpd.GeoDataFrame:
        data = [{"town": town.name, **service.to_dict()} for town in self.towns for service in town.services]
        return gpd.GeoDataFrame(data, crs=self.crs)

    def get_towns_gdf(self) -> gpd.GeoDataFrame:
        data = [town.to_dict() for town in self.towns]
        gdf = gpd.GeoDataFrame(data, crs=self.crs).rename(columns={"name": "town_name"}).set_index("id", drop=True)
        for service_type in self.service_types:
            capacity_column = f"capacity_{service_type.name}"
            count_column = f"count_{service_type.name}"
            if not capacity_column in gdf.columns:
                gdf[capacity_column] = 0
            if not count_column in gdf.columns:
                gdf[count_column] = 0
        gdf = gdf.sjoin(
            self.settlements[["geometry", "name"]].rename(columns={"name": "settlement_name"}),
            how="left",
            predicate="within",
            lsuffix="town",
            rsuffix="settlement",
        )
        gdf = gdf.sjoin(
            self.districts[["geometry", "name"]].rename(columns={"name": "district_name"}),
            how="left",
            predicate="within",
            lsuffix="town",
            rsuffix="district",
        )
        return gdf.drop(columns=["index_settlement", "index_district"]).fillna(0)

    def get_service_types_df(self) -> pd.DataFrame:
        data = [service_type.to_dict() for service_type in self.service_types]
        return pd.DataFrame(data)

    @singledispatchmethod
    def __getitem__(self, arg):
        raise NotImplementedError(f"Can't access object with such argument type {type(arg)}")

    # Make city_model subscriptable, to access block via ID like city_model[123]
    @__getitem__.register(int)
    def _(self, town_id):
        if not town_id in self._towns:
            raise KeyError(f"Can't find town with such id: {town_id}")
        return self._towns[town_id]

    # Make city_model subscriptable, to access service type via name like city_model['schools']
    @__getitem__.register(str)
    def _(self, service_type_name):
        if not service_type_name in self._service_types:
            raise KeyError(f"Can't find service type with such name: {service_type_name}")
        return self._service_types[service_type_name]

    @__getitem__.register(tuple)
    def _(self, towns):
        (town_a, town_b) = towns
        if isinstance(town_a, Town):
            town_a = town_a.id
        if isinstance(town_b, Town):
            town_b = town_b.id
        return self.accessibility_matrix.loc[town_a, town_b]

    @staticmethod
    def from_pickle(file_path: str):
        """Load region model from a .pickle file"""
        state = None
        with open(file_path, "rb") as f:
            state = pickle.load(f)
        return state

    def to_pickle(self, file_path: str):
        """Save region model to a .pickle file"""
        with open(file_path, "wb") as f:
            pickle.dump(self, f)
